from uuid import uuid4

import jwt
from django.utils import timezone
from django.utils.functional import cached_property
from jwt import InvalidTokenError

from ievv_auth.ievv_jwt.exceptions import JWTBackendError
from ievv_auth.ievv_jwt.utils.load_settings import load_settings

ALLOWED_ALGORITHMS = [
    'HS256',
    'HS384',
    'HS512',
    'RS256',
    'RS384',
    'RS512',
]


class AbstractBackend:

    @classmethod
    def get_backend_name(cls):
        raise NotImplementedError('Please implement get_backend_name')

    @cached_property
    def settings(self):
        return load_settings(self.__class__.get_backend_name())

    @property
    def algorithm(self):
        algorithm = self.settings['ALGORITHM']
        if algorithm not in ALLOWED_ALGORITHMS:
            raise JWTBackendError(f'{self.algorithm} is not one of the allowed algorithms')
        return algorithm

    @property
    def signing_key(self):
        return self.settings['SIGNING_KEY']

    @property
    def audience(self):
        return self.settings.get('AUDIENCE', None)

    @property
    def issuer(self):
        return self.settings.get('ISSUER', None)

    @property
    def subject(self):
        return self.settings.get('SUBJECT', None)

    @property
    def verifying_key(self):
        if self.algorithm.startswith('HS'):
            return self.signing_key
        if not self.settings['VERIFYING_KEY']:
            raise JWTBackendError(f'Verifying key cannot be None when algorithm is {self.algorithm}')
        return self.settings['VERIFYING_KEY']

    @property
    def access_token_expiration(self):
        return timezone.now() + self.settings['ACCESS_TOKEN_LIFETIME']

    @property
    def jti(self):
        return uuid4().hex

    def make_payload(self, base_payload=None):
        if not base_payload:
            payload = {}
        else:
            payload = base_payload
        if self.audience:
            payload['aud'] = self.audience

        if self.issuer:
            payload['iss'] = self.issuer

        if self.subject:
            payload['sub'] = self.subject

        payload['exp'] = self.access_token_expiration
        payload['iat'] = timezone.now()
        payload[self.settings['JTI_CLAIM']] = self.jti
        payload['jwt_backend_name'] = self.__class__.get_backend_name()
        return payload

    def encode(self, base_payload=None):
        token = jwt.encode(
            self.make_payload(base_payload=base_payload),
            self.signing_key,
            algorithm=self.algorithm
        )
        return token.decode('utf-8')

    def decode(self, token, verify=True):
        try:
            return jwt.decode(token, self.verifying_key, algorithms=[self.algorithm], verify=verify,
                              audience=self.audience, issuer=self.issuer,
                              options={'verify_aud': self.audience is not None})
        except InvalidTokenError:
            raise JWTBackendError('Token is invalid or expired')

    def make_authenticate_success_response(self, *args, **kwargs):
        return {}

    @classmethod
    def make_instance_from_raw_jwt(cls, raw_jwt):
        return cls()


class BaseBackend(AbstractBackend):
    @classmethod
    def get_backend_name(cls):
        return 'default'



from ievv_auth.project.develop.common_settings import *

# Disable migrations when running tests
class DisableMigrations(object):

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        import django
        if django.VERSION[0:3] > (1, 11, 0):
            return None
        return 'notmigrations'

MIGRATION_MODULES = DisableMigrations()



from ievv_auth.project.default.settings import *


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "dbdev.sqlite3",
    }
}
DATABASES['default']['PORT'] = 27253



import base64
import json
from calendar import timegm
from unittest import mock
from unittest.mock import PropertyMock

import jwt as py_jwt
from django.test import TestCase
from django.utils import timezone
from model_bakery import baker

from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
from ievv_auth.ievv_jwt.exceptions import JWTBackendError


class TestApiKeyBackend(TestCase):

    def test_sanity(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertIn('iat', decoded)
        self.assertIn('jti', decoded)
        self.assertEqual(decoded['api_key_id'], api_key.id)
        self.assertEqual(decoded['scope'], ['read', 'write'])

    def test_fields_which_is_not_overridable_is_not_changed(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'exp': 123,
                'iat': 123,
                'jti': 123
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertNotEqual(decoded['exp'], 123)
        self.assertIn('iat', decoded)
        self.assertNotEqual(decoded['iat'], 123)
        self.assertIn('jti', decoded)
        self.assertNotEqual(decoded['jti'], 123)
        self.assertEqual(decoded['api_key_id'], api_key.id)

    def test_verify_intercepted_payload_extend_expiration(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['exp'] = timegm((timezone.now() + timezone.timedelta(weeks=200)).utctimetuple())
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_verify_intercepted_payload_added_additional_scope(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['scope'] = 'admin'
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_sign_jwt_with_another_secret(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        decoded = backend.decode(jwt)
        new_jwt = py_jwt.encode(payload=decoded, key='asdxxxxxxxxxxxxxxxxxxxxxxxxxxx')
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'ACCESS_TOKEN_LIFETIME': timezone.timedelta(minutes=0),
            }
        }):
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.api_key_backend.ApiKeyBackend.access_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                api_key = baker.make(
                    'ievv_api_key.ScopedApiKey',
                    base_jwt_payload={
                        'scope': ['read', 'write']
                    }
                )
                backend = ApiKeyBackend()
                jwt = backend.encode(base_payload={
                    **api_key.base_jwt_payload,
                    'api_key_id': api_key.id
                })
                with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
                    backend.decode(jwt, verify=True)

    def test_make_instance_from_raw_jwt(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        jwt = backend.encode(base_payload={
            **api_key.base_jwt_payload,
            'api_key_id': api_key.id
        })
        backend_instance = ApiKeyBackend.make_instance_from_raw_jwt(raw_jwt=jwt)
        self.assertIsInstance(backend_instance, ApiKeyBackend)

    def test_make_authenticate_success_response(self):
        with mock.patch(
            'ievv_auth.ievv_jwt.backends.api_key_backend.ApiKeyBackend.encode',
            return_value='test'
        ):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                }
            )
            backend = ApiKeyBackend()
            self.assertDictEqual(
                backend.make_authenticate_success_response(
                    base_payload={
                        **api_key.base_jwt_payload
                    }),
                {'access': 'test'}
            )



from django.conf import settings
from django.utils import timezone


DEFAULT_SETTINGS = {
    'ACCESS_TOKEN_LIFETIME': timezone.timedelta(minutes=2),
    'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
    'BLACKLIST_AFTER_ROTATION': True,

    'ALGORITHM': 'HS256',
    'SIGNING_KEY': settings.SECRET_KEY,
    'VERIFYING_KEY': None,
    'AUDIENCE': None,
    'ISSUER': None,

    'AUTH_TOKEN_CLASSES': ('rest_framework_simplejwt.tokens.AccessToken',),
    'TOKEN_TYPE_CLAIM': 'token_type',

    'JTI_CLAIM': 'jti',

    'SLIDING_TOKEN_REFRESH_EXP_CLAIM': 'refresh_exp',
    'SLIDING_TOKEN_LIFETIME': timezone.timedelta(minutes=5),
    'SLIDING_TOKEN_REFRESH_LIFETIME': timezone.timedelta(days=1),
}

GLOBAL_SETTINGS = {
    'AUTH_HEADER_TYPES': ('Bearer',)
}



import jwt
from ievv_opensource.utils.singleton import Singleton


class JWTBackendRegistry(Singleton):

    def __init__(self):
        super(JWTBackendRegistry, self).__init__()
        self._backend_map = {}

    def set_backend(self, backend):
        self._backend_map[backend.get_backend_name()] = backend

    def get_backend(self, backend_name):
        return self._backend_map.get(backend_name, None)

    def get_backend_choices(self):
        choices = []
        for key in self._backend_map:
            choices.append((key, key))
        return choices


def get_backend_from_raw_jwt(raw_jwt):
    payload = jwt.decode(jwt=raw_jwt, verify=False)
    return JWTBackendRegistry.get_instance().get_backend(payload['jwt_backend_name'])



import re
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.db import models
from django.utils.translation import gettext_lazy
from django.contrib.auth.hashers import check_password, make_password
from django.utils.crypto import get_random_string


class BaseAPIKeyManager(models.Manager):

    def make_prefix(self):
        """
        Generates a unique api key prefix

        Returns:
            String: generated prefix for api key
        """
        while True:
            prefix = get_random_string(length=8)
            if not self.filter(prefix=prefix).exists():
                break
        return prefix

    def make_secret_key(self):
        """
        Generates a unique api key secret

        Returns:
             String: generated secret for api key
        """
        return get_random_string(length=getattr(settings, 'IEVV_API_KEY_SECRET_KEY_LENGTH', 32))

    def hash_secret_key(self, secret_key):
        """
        Hash secret key

        Args:
            secret_key: The secret key which will be hashed

        Returns:
             String: Hashed secret key
        """
        return make_password(secret_key)

    def make_expiration_datetime(self):
        """
        makes expiration datetime for the api key

        Returns:
             Datetime: Expiration datetime
        """
        return timezone.now() + timezone.timedelta(days=getattr(settings, 'IEVV_API_KEY_EXPIRATION_DAYS', 180))

    def generate_api_key(self):
        """
        Generate api key

        Returns:
            String: api_key (prefix).(secret)
            String: The api key prefix
            String: hashed_key the hashed secret
        """
        prefix = self.make_prefix()
        secret_key = self.make_secret_key()
        hashed_key = self.hash_secret_key(secret_key=secret_key)
        api_key = '{}.{}'.format(prefix, secret_key)
        return api_key, prefix, hashed_key

    def create_key(self, **kwargs):
        api_key, prefix, hashed_key = self.generate_api_key()
        instance = self.model(
            prefix=prefix,
            hashed_key=hashed_key,
            expiration_datetime=self.make_expiration_datetime(),
            **kwargs
        )
        instance.full_clean()
        instance.save()
        return api_key, instance

    def _is_valid(self, api_key):
        #: Validate format for the api key
        if len(api_key.split('.')) != 2:
            return None, False, None

        [prefix, secret] = api_key.split('.')
        try:
            instance = self.get(prefix=prefix)
        except self.model.DoesNotExist:
            return None, False, None
        valid = True
        message = 'Authentication successful'
        if not instance.verify(secret_key=secret):
            valid = False
            message = 'Authentication failed, secret key is not valid'
        if instance.revoked:
            valid = False
            message = 'Authentication failed, key has been revoked'
        if instance.has_expired:
            valid = False
            message = 'Authentication failed, key has expired'
        return instance, valid, message

    def is_valid(self, api_key):
        """
        Simply checks whether the api key is valid
        Returns True if the API Key is valid, False otherwise.

        It will check:
            - if api key prefix exists
            - if the api key has expired
            - if the hashed secret matches the one on database

        Args:
            api_key: The api key

        Returns:
            Boolean: If the api_key is valid
            Object: api key instance
        """
        instance, valid, _ = self._is_valid(api_key=api_key)
        return valid, instance

    def is_valid_with_logging(self, api_key, extra=None):
        """
        Returns True if the API Key is valid, False otherwise.
        Some logging is appended for the authentication try

        It will check:
            - if api key prefix exists
            - if the api key has expired
            - if the hashed secret matches the one on database

        Args:
            api_key: The api key
            extra (optional): optional dict for logging entry

        Returns:
            Boolean: If the api_key is valid
            Object: api key instance
        """
        if not extra:
            extra = {}
        instance, valid, message = self._is_valid(api_key=api_key)

        if instance:
            extra.update({'message': message})
            instance.log_authentication_attempt(extra)
        return valid, instance


class AbstractAPIKey(models.Model):
    objects = BaseAPIKeyManager()

    #: api key prefix (prefix).(secret)
    prefix = models.CharField(
        max_length=8,
        unique=True,
        editable=False
    )

    #: hashed key
    hashed_key = models.CharField(
        max_length=100,
        editable=False,
    )

    #: free text field
    name = models.CharField(
        max_length=255,
        blank=True,
        default=None,
        null=True,
        verbose_name=gettext_lazy('Name'),
        help_text=gettext_lazy('Free text for the api key')
    )

    #: If the api key has been revoked
    revoked = models.BooleanField(
        default=False,
        verbose_name=gettext_lazy('Revoked'),
        help_text=gettext_lazy('If the api key has been revoked, once revoked it cannot be undone')
    )

    #: created datetime
    created = models.DateTimeField(
        auto_now_add=True,
        editable=False,
        verbose_name=gettext_lazy('Created datetime'),
        help_text=gettext_lazy('When the api key was created')
    )

    #: datetime when the key expires
    expiration_datetime = models.DateTimeField(
        blank=False,
        null=False,
        verbose_name=gettext_lazy('Expires'),
        help_text=gettext_lazy('The datetime when the token expires'),
        editable=False
    )

    def __init__(self, *args, **kwargs):
        super(AbstractAPIKey, self).__init__(*args, **kwargs)
        self._initial_revoked = self.revoked

    class Meta:
        abstract = True
        verbose_name = gettext_lazy('API key')
        verbose_name_plural = gettext_lazy('API keys')

    def log_authentication_attempt(self, log_data):
        raise NotImplementedError('Please implement logging_model_class')

    @property
    def has_expired(self):
        return self.expiration_datetime < timezone.now()

    def verify(self, secret_key):
        return check_password(secret_key, self.hashed_key)

    def _validate_revoked(self):
        if self._initial_revoked and not self.revoked:
            raise ValidationError(
                gettext_lazy('Cannot set revoked to False once revoked')
            )

    def clean(self):
        self._validate_revoked()


class ScopedAPIKey(AbstractAPIKey):

    #: Backend type
    jwt_backend_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        default=None,
    )

    #: Should contain the payload of the jwt issued after authentication
    base_jwt_payload = models.JSONField(
        default=dict,
        null=False,
        blank=True
    )

    class Meta:
        verbose_name = gettext_lazy('Scoped api key')
        verbose_name_plural = gettext_lazy('Scoped api keys')

    def log_authentication_attempt(self, log_data):
        ScopedApiKeyAuthenticationLog.objects.create(api_key=self, log_data=log_data)

    def __str__(self):
        return f'{self.name} - id<{self.id}>'


class AbstractAuthenticationLog(models.Model):
    #: Log data contains authentication attempt message and auxiliary data
    log_data = models.JSONField(null=True, blank=False, default=dict)

    #: created datetime
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class ScopedApiKeyAuthenticationLog(AbstractAuthenticationLog):
    #: The instance of ScopedAPIKey this log belongs to.
    api_key = models.ForeignKey(to=ScopedAPIKey, on_delete=models.CASCADE)



from unittest.mock import patch

from django import test
from django.core.exceptions import ValidationError
from django.utils import timezone

from ievv_auth.ievv_api_key.models import ScopedAPIKey, ScopedApiKeyAuthenticationLog


class TestScopedApiKey(test.TestCase):

    def test_valid_api_key(self):
        api_key, instance = ScopedAPIKey.objects.create_key()
        is_valid, instance = ScopedAPIKey.objects.is_valid(api_key=api_key)
        self.assertTrue(is_valid)

    def test_api_key_revoked(self):
        api_key, instance = ScopedAPIKey.objects.create_key()
        instance.revoked = True
        instance.save()
        is_valid, instance = ScopedAPIKey.objects.is_valid(api_key=api_key)
        self.assertFalse(is_valid)

    def test_api_key_expired(self):
        with patch.object(ScopedAPIKey.objects, 'make_expiration_datetime', return_value=timezone.now() - timezone.timedelta(days=1)):
            api_key, instance = ScopedAPIKey.objects.create_key()
            is_valid, instance = ScopedAPIKey.objects.is_valid(api_key=api_key)
            self.assertFalse(is_valid)

    def test_api_key_is_valid_api_key_does_not_exists(self):
        is_valid, instance = ScopedAPIKey.objects.is_valid(api_key='XXXXXXXX.YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY')
        self.assertFalse(is_valid)

    def test_valid_api_key_with_logging(self):
        api_key, instance = ScopedAPIKey.objects.create_key()
        is_valid, instance = ScopedAPIKey.objects.is_valid_with_logging(api_key=api_key)
        self.assertTrue(is_valid)
        instance.refresh_from_db()
        self.assertEqual(ScopedApiKeyAuthenticationLog.objects.count(), 1)
        log = ScopedApiKeyAuthenticationLog.objects.first()
        self.assertEqual(log.api_key.id, instance.id)
        self.assertEqual(log.log_data.get('message'), 'Authentication successful')

    def test_api_key_revoked_with_logging(self):
        api_key, instance = ScopedAPIKey.objects.create_key()
        instance.revoked = True
        instance.save()
        is_valid, instance = ScopedAPIKey.objects.is_valid_with_logging(api_key=api_key)
        self.assertFalse(is_valid)
        self.assertEqual(ScopedApiKeyAuthenticationLog.objects.count(), 1)
        log = ScopedApiKeyAuthenticationLog.objects.first()
        self.assertEqual(log.api_key.id, instance.id)
        self.assertEqual(log.log_data.get('message'), 'Authentication failed, key has been revoked')

    def test_api_key_expired_with_logging(self):
        with patch.object(ScopedAPIKey.objects, 'make_expiration_datetime',
                          return_value=timezone.now() - timezone.timedelta(days=1)):
            api_key, instance = ScopedAPIKey.objects.create_key()
            is_valid, instance = ScopedAPIKey.objects.is_valid_with_logging(api_key=api_key)
            self.assertFalse(is_valid)
            self.assertEqual(ScopedApiKeyAuthenticationLog.objects.count(), 1)
            log = ScopedApiKeyAuthenticationLog.objects.first()
            self.assertEqual(log.api_key.id, instance.id)
            self.assertEqual(log.log_data.get('message'), 'Authentication failed, key has expired')

    def test_should_not_be_able_to_set_revoked_to_false_after_being_true(self):
        api_key, instance = ScopedAPIKey.objects.create_key()
        instance.revoked = True
        instance.save()
        [prefix, _] = api_key.split('.')
        instance = ScopedAPIKey.objects.get(prefix=prefix)
        with self.assertRaisesMessage(ValidationError, 'Cannot set revoked to False once revoked'):
            instance.revoked = False
            instance.full_clean()
